// David Pilling 6th June 2014
// Receive Oregon v2 packets and print out values
// Also receives packets from my own liquid depth sensor

// the conversion from raw packet to numeric values was inspired by
// http://www.disk91.com/2013/technology/hardware/oregon-scientific-sensors-with-raspberry-pi/
// I suspect none of that code remains though


// I made two key changes to the code below
// 1) make preamble 24 bits (as others have done)
// 2) packet length 159 bits - one of my sensors does not seem to send 160 bits

// Oregon V2 decoder added - Dominique Pierre
// Oregon V3 decoder revisited - Dominique Pierre
// New code to decode OOK signals from weather sensors, etc.
// 2010-04-11 <jcw@equi4.com> http://opensource.org/licenses/mit-license.php
// $Id: ookDecoder.pde 5331 2010-04-17 10:45:17Z jcw $

class DecodeOOK {
protected:
    byte total_bits, bits, flip, state, pos, data[25];

    virtual char decode (word width) =0;
    
public:

    enum { UNKNOWN, T0, T1, T2, T3, OK, DONE };

    DecodeOOK () { resetDecoder(); }

    bool nextPulse (word width) {
        if (state != DONE)
        
            switch (decode(width)) {
                case -1: resetDecoder(); break;
                case 1:  done(); break;
            }
        return isDone();
    }
    
    bool isDone () const { return state == DONE; }

    const byte* getData (byte& count) const {
        count = pos;
        return data; 
    }
    
    void resetDecoder () {
        total_bits = bits = pos = flip = 0;
        state = UNKNOWN;
    }
    
    // add one bit to the packet data buffer
    
    virtual void gotBit (char value) {
        total_bits++;
        byte *ptr = data + pos;
        *ptr = (*ptr >> 1) | (value << 7);

        if (++bits >= 8) {
            bits = 0;
            if (++pos >= sizeof data) {
                resetDecoder();
                return;
            }
        }
        state = OK;
    }
    
    // store a bit using Manchester encoding
    void manchester (char value) {
        flip ^= value; // manchester code, long pulse flips the bit
        gotBit(flip);
    }
    
    // move bits to the front so that all the bits are aligned to the end
    void alignTail (byte max =0) {
        // align bits
        if (bits != 0) {
            data[pos] >>= 8 - bits;
            for (byte i = 0; i < pos; ++i)
                data[i] = (data[i] >> bits) | (data[i+1] << (8 - bits));
            bits = 0;
        }
        // optionally shift bytes down if there are too many of 'em
        if (max > 0 && pos > max) {
            byte n = pos - max;
            pos = max;
            for (byte i = 0; i < pos; ++i)
                data[i] = data[i+n];
        }
    }
    
    void reverseBits () {
        for (byte i = 0; i < pos; ++i) {
            byte b = data[i];
            for (byte j = 0; j < 8; ++j) {
                data[i] = (data[i] << 1) | (b & 1);
                b >>= 1;
            }
        }
    }
    
    void reverseNibbles () {
        for (byte i = 0; i < pos; ++i)
            data[i] = (data[i] << 4) | (data[i] >> 4);
    }
    
    void done () {
        while (bits)
            gotBit(0); // padding
        state = DONE;
    }
};

// 433 MHz decoders


class OregonDecoderV2 : public DecodeOOK {
public:
    OregonDecoderV2() {}
    
    // add one bit to the packet data buffer
    virtual void gotBit (char value) {
        if(!(total_bits & 0x01))
        {
            data[pos] = (data[pos] >> 1) | (value ? 0x80 : 00);
        }
        total_bits++;
        pos = total_bits >> 4;
        if (pos >= sizeof data) {
            resetDecoder();
            return;
        }
        state = OK;
    }
    
    virtual char decode (word width) {
        if (200 <= width && width < 1200) {
            byte w = width >= 700;
            switch (state) {
                case UNKNOWN:
                    if (w != 0) {
                        // Long pulse
                        ++flip;
                    } else if (24 <= flip) {
                        // Short pulse, start bit
                        flip = 0;
                        state = T0;
                    } else {
                      // Reset decoder
                        return -1;
                    }
                    break;
                case OK:
                    if (w == 0) {
                        // Short pulse
                        state = T0;
                    } else {
                        // Long pulse
                        manchester(1);
                    }
                    break;
                case T0:
                    if (w == 0) {
                      // Second short pulse
                        manchester(0);
                    } else {
                        // Reset decoder
                        return -1;
                    }
                    break;
            }
        } else {
            return -1;
        }
        return total_bits == 159 ? 1: 0;
    }
};


OregonDecoderV2 orscV2;



volatile word pulse;


void ext_int_1(void) {
    static word last;
    // determine the pulse length in microseconds, for either polarity
    pulse = micros() - last;
    last += pulse;
}


char itox(int n)
{
 const char array[]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
 return(array[n]); 
}  


void reportSerial (const char* s, class DecodeOOK& decoder)
{
 byte pos;
 char string[64];
 char * p=string;

 const byte* data = decoder.getData(pos);

 strcpy(string,s);
 p+=strlen(s);

 *p++=' ';
 
 for(byte i = 0; i < pos; ++i) 
 {
  *p++=itox(data[i] >> 4);
  *p++=itox(data[i] & 0x0F);
 }
    
 *p=0;
 
 Serial.println(string);
    
 decoder.resetDecoder();
 
 char * pt=string+5;
 char temp[8];
 char humid[4];
 
 temp[0] = pt[10] ; temp[1] = pt[11] ; temp[2] = pt[8] ; temp[3] = '\0';
 humid[0] = pt[15] ; humid[1] = pt[12]; humid[2] = '0' ; humid[3] = '\0';

 if(strncmp(pt,"1A2D",4)==0)
 {
  Serial.print("temp ");
  Serial.println(temp);
 
  Serial.print("humid ");
  Serial.println(humid);
 }
 else
 if(strncmp(pt,"EA4C",4)==0) 
 {
  Serial.print("temp ");
  Serial.println(temp);    
 } 
 else
 if(strncmp(pt,"1ABC",4)==0) 
 {
  // 1ABC2053512300002D00
 
  strncpy(temp,pt+6,2);
  temp[2]=0; 
  Serial.print("vcc ");
  Serial.print(temp);
  
  strncpy(temp,pt+8,2); 
  temp[2]=0; 
  Serial.print(" humid ");
  Serial.print(temp);
  
  strncpy(temp,pt+10,2); 
  temp[2]=0; 
  Serial.print(" temp ");
  Serial.print(temp);
  
  strncpy(temp,pt+12,4);
  temp[4]=0; 
  Serial.print(" freq ");
  Serial.println(temp);

 }
}


void setup () {
    Serial.begin(9600);
    Serial.println("\n[ookDecoder]");
    

   attachInterrupt(1, ext_int_1, CHANGE);

}

void loop () {
    static int i = 0;
    cli();
    word p = pulse;
    
    pulse = 0;
    sei();

   
    if (p != 0) {
        if (orscV2.nextPulse(p))
            reportSerial("OSV2", orscV2);  
           
}
}

