// David Pilling 6th June 2014
// code for my liquid level sensor



#include "LowPower.h"
#include <FreqCounter.h>


#define pin_led 13
#define pin_sensor_power 12
#define pin_sensor_signal 5
#define pin_temp_power 10
#define pin_temp_signal 9
#define pin_tx_power 8
#define pin_tx_data (7)

int mode_speed=1;  // bigger == slower
int mode_slowspeed=4;
int mode_count=0;
int mode_switch=100; // after mode_count equals this switch to slow mode
int mode_slow=0;     // true when in slow mode





/*
 * connectingStuff, Oregon Scientific v2.1 Emitter
 * http://connectingstuff.net/blog/encodage-protocoles-oregon-scientific-sur-arduino/
 *
 * Copyright (C) 2013 olivier.lebrun@gmail.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


void writeBCD(byte * p,int nybble,int width,long value)
{
 char buffer[16]; 
 char fmt[8];
 int  byteindex=nybble/2;
 int  i;
 int  c;

 nybble=nybble-byteindex*2; 
  
  
// Serial.println(value); 

 sprintf(fmt,"%%0%dd",width); 
 sprintf(buffer,fmt,value);

// Serial.println(buffer);


 for(i=0;i<width;i++)
 {
  c=buffer[i]-'0';
//   Serial.println(c);
  if(nybble & 0x1)
  {
   p[byteindex]=(p[byteindex] & 0xF0) | c;
   nybble=0;
   byteindex++;
  }
  else
  {
   p[byteindex]=(p[byteindex] & 0xF) | (c<<4);
   nybble++;
  } 
 } 
}





const unsigned long TIME = 512;
const unsigned long TWOTIME = 1024;
 
#define SEND_HIGH() digitalWrite(pin_tx_data, 1)
#define SEND_LOW() digitalWrite(pin_tx_data, 0)



 
// Buffer for Oregon message
  byte OregonMessageBuffer[9];

 
/**
 * \brief    Send logical "0" over RF
 * \details  azero bit be represented by an off-to-on transition
 * \         of the RF signal at the middle of a clock period.
 * \         Remenber, the Oregon v2.1 protocol add an inverted bit first 
 */
inline void sendZero(void) 
{
  SEND_HIGH();
  delayMicroseconds(TIME);
//  pulse(TIME,HIGH);
  SEND_LOW();
  delayMicroseconds(TWOTIME);
//  pulse(TWOTIME,LOW);
  SEND_HIGH();
  delayMicroseconds(TIME);
//  pulse(TIME,HIGH);  
}
 
/**
 * \brief    Send logical "1" over RF
 * \details  a one bit be represented by an on-to-off transition
 * \         of the RF signal at the middle of a clock period.
 * \         Remenber, the Oregon v2.1 protocol add an inverted bit first 
 */
inline void sendOne(void) 
{
   SEND_LOW();
   delayMicroseconds(TIME);
//   pulse(TIME,LOW);   
   SEND_HIGH();
   delayMicroseconds(TWOTIME);
//   pulse(TWOTIME,HIGH);   
   SEND_LOW();
   delayMicroseconds(TIME);
//   pulse(TIME,LOW);   
}
 
/**
* Send a bits quarter (4 bits = MSB from 8 bits value) over RF
*
* @param data Source data to process and sent
*/
 
/**
 * \brief    Send a bits quarter (4 bits = MSB from 8 bits value) over RF
 * \param    data   Data to send
 */
inline void sendQuarterMSB(const byte data) 
{
  (bitRead(data, 4)) ? sendOne() : sendZero();
  (bitRead(data, 5)) ? sendOne() : sendZero();
  (bitRead(data, 6)) ? sendOne() : sendZero();
  (bitRead(data, 7)) ? sendOne() : sendZero();
}
 
/**
 * \brief    Send a bits quarter (4 bits = LSB from 8 bits value) over RF
 * \param    data   Data to send
 */
inline void sendQuarterLSB(const byte data) 
{
  (bitRead(data, 0)) ? sendOne() : sendZero();
  (bitRead(data, 1)) ? sendOne() : sendZero();
  (bitRead(data, 2)) ? sendOne() : sendZero();
  (bitRead(data, 3)) ? sendOne() : sendZero();
}
 
/******************************************************************/
/******************************************************************/
/******************************************************************/
 
/**
 * \brief    Send a buffer over RF
 * \param    data   Data to send
 * \param    size   size of data to send
 */
void sendData(byte *data, byte size)
{
  for(byte i = 0; i < size; ++i)
  {
    sendQuarterLSB(data[i]);
    sendQuarterMSB(data[i]);
  }
}
 
/**
 * \brief    Send an Oregon message
 * \param    data   The Oregon message
 */
void sendOregon(byte *data, byte size)
{
    sendPreamble();
    //sendSync();
    sendData(data, size);
    sendPostamble();
}
 
/**
 * \brief    Send preamble
 * \details  The preamble consists of 16 "1" bits
 */
inline void sendPreamble(void)
{
  byte PREAMBLE[]={0xFF,0xFF};
  sendData(PREAMBLE, 2);
}
 
/**
 * \brief    Send postamble
 * \details  The postamble consists of 8 "0" bits
 */
inline void sendPostamble(void)
{
  byte POSTAMBLE[]={0x00};
  sendData(POSTAMBLE, 1);  

}
 
/**
 * \brief    Send sync nibble
 * \details  The sync is 0xA. It is not use in this version since the sync nibble
 * \         is include in the Oregon message to send.
 */
inline void sendSync(void)
{
  sendQuarterLSB(0xA);
}
 
/******************************************************************/
/******************************************************************/
/******************************************************************/
 
/**
 * \brief    Set the sensor type
 * \param    data       Oregon message
 * \param    type       Sensor type
 */
inline void setType(byte *data, byte* type) 
{
  data[0] = type[0];
  data[1] = type[1];
}
 
/**
 * \brief    Set the sensor channel
 * \param    data       Oregon message
 * \param    channel    Sensor channel (0x10, 0x20, 0x30)
 */
inline void setChannel(byte *data, byte channel) 
{
    data[2] = channel;
}
 
/**
 * \brief    Set the sensor ID
 * \param    data       Oregon message
 * \param    ID         Sensor unique ID
 */
inline void setId(byte *data, byte ID) 
{
  data[3] = ID;
}
 
/**
 * \brief    Set the sensor battery level
 * \param    data       Oregon message
 * \param    level      Battery level (0 = low, 1 = high)
 */
void setBatteryLevel(byte *data, byte level)
{
  if(!level) data[4] = 0x0C;
  else data[4] = 0x00;
}
 
/**
 * \brief    Set the sensor temperature
 * \param    data       Oregon message
 * \param    temp       the temperature
 */
void setTemperature(byte *data, float temp) 
{
  // Set temperature sign
  if(temp < 0)
  {
    data[6] = 0x08;
    temp *= -1;  
  }
  else
  {
    data[6] = 0x00;
  }
 
  // Determine decimal and float part
  int tempInt = (int)temp;
  int td = (int)(tempInt / 10);
  int tf = (int)round((float)((float)tempInt/10 - (float)td) * 10);
 
  int tempFloat =  (int)round((float)(temp - (float)tempInt) * 10);
 
  // Set temperature decimal part
  data[5] = (td << 4);
  data[5] |= tf;
 
  // Set temperature float part
  data[4] |= (tempFloat << 4);
}
 
/**
 * \brief    Set the sensor humidity
 * \param    data       Oregon message
 * \param    hum        the humidity
 */
void setHumidity(byte* data, byte hum)
{
    data[7] = (hum/10);
    data[6] |= (hum - data[7]*10) << 4;
}
 
/**
 * \brief    Sum data for checksum
 * \param    count      number of bit to sum
 * \param    data       Oregon message
 */
int Sum(byte count, const byte* data)
{
  int s = 0;
 
  for(byte i = 0; i<count;i++)
  {
    s += (data[i]&0xF0) >> 4;
    s += (data[i]&0xF);
  }
 
  if(int(count) != count)
    s += (data[count]&0xF0) >> 4;
 
  return s;
}
 
/**
 * \brief    Calculate checksum
 * \param    data       Oregon message
 */
void calculateAndSetChecksum(byte* data)
{
 data[8] = ((Sum(8, data) - 0xa) & 0xFF);
}

void transmitter_setup(void)
{
  byte ID[] = {0x1A,0xBC};
  
  pinMode(pin_tx_data , OUTPUT);
  SEND_LOW();  
//  pulse(0,LOW);
 
  setType(OregonMessageBuffer, ID);
  setChannel(OregonMessageBuffer, 0x20);
  setId(OregonMessageBuffer, 0xBB);
  
}


void transmitter_send(void)
{
  calculateAndSetChecksum(OregonMessageBuffer);
 
  Serial.println("send message");
  // Show the Oregon Message
  for (byte i = 0; i < sizeof(OregonMessageBuffer); ++i)   
  {
    Serial.print(OregonMessageBuffer[i] >> 4, HEX);
    Serial.print(OregonMessageBuffer[i] & 0x0F, HEX);
  }
 
  Serial.println();
  // Send the Message over RF
  sendOregon(OregonMessageBuffer, sizeof(OregonMessageBuffer));
  // Send a "pause"
  SEND_LOW();
//  pulse(0,LOW);

  delayMicroseconds(TWOTIME*8);
  // Send a copie of the first message. The v2.1 protocol send the
  // message two time 
  sendOregon(OregonMessageBuffer, sizeof(OregonMessageBuffer));
 
  // Wait for 30 seconds before send a new message 
  SEND_LOW();  
 
}




// https://code.google.com/p/tinkerit/wiki/SecretThermometer

long readTemp() {
  long result;
  // Read temperature sensor against 1.1V reference
  ADMUX = _BV(REFS1) | _BV(REFS0) | _BV(MUX3);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = (result - 125) * 1075;
  return result;
}


// https://code.google.com/p/tinkerit/wiki/SecretVoltmeter

long readVcc() {
  long result;
  // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = 1126400L / result; // Back-calculate AVcc in mV
  return result;
}



#include <dht11.h>

void powerup_temp(void)
{
 pinMode(pin_temp_power, OUTPUT);  
 digitalWrite(pin_temp_power, HIGH);
}


void powerdown_temp(void)
{
 pinMode(pin_temp_signal, INPUT);   
 digitalWrite(pin_temp_signal, LOW);

 digitalWrite(pin_temp_power, LOW);
 pinMode(pin_temp_power, INPUT);   
} 


void read_temp(void)
{
 dht11 DHT11;

 
 int chk = DHT11.read(pin_temp_signal);
 
// Serial.print("Read sensor: ");
 switch (chk)
 {
    case DHTLIB_OK: 
		Serial.println("OK"); 
		break;
    case DHTLIB_ERROR_CHECKSUM: 
		Serial.println("Checksum error"); 
		break;
    case DHTLIB_ERROR_TIMEOUT: 
		Serial.println("Time out error"); 
		break;
    default: 
		Serial.println("Unknown error"); 
		break;
 }

// Serial.print("Humidity (%): ");
// Serial.println((float)DHT11.humidity, 2);

// Serial.print("Temperature (C): ");
// Serial.println((float)DHT11.temperature, 2);
 


 writeBCD(OregonMessageBuffer,8,2,DHT11.humidity);
 writeBCD(OregonMessageBuffer,10,2,DHT11.temperature);

 powerdown_temp();
}




void powerup_sensor(void)
{
 pinMode(pin_sensor_power, OUTPUT);  
 digitalWrite(pin_sensor_power, HIGH);
  
  
}  


void powerdown_sensor(void)
{
  
 digitalWrite(pin_sensor_power, LOW);
 pinMode(pin_sensor_power, INPUT);   
  
}  



void read_sensor(void)
{
 long frq;
  
 FreqCounter::f_comp=10;   // Cal Value / Calibrate with professional Freq Counter
 FreqCounter::start(100);  // 100 ms Gate Time

 while(FreqCounter::f_ready == 0); 
// Serial.print("Freq ready");
// Serial.println(FreqCounter::f_ready,DEC);

 frq=FreqCounter::f_freq/10; 
 if(frq>9999) frq=9999;

// Serial.print("Freq ");
// Serial.println(frq,DEC);
 
 writeBCD(OregonMessageBuffer,12,4,frq);
 
 powerdown_sensor();
}  


// voltage 2 5.2
// temp    2 10.0
// hum     2 66
// freq    4
// 3,4,5,6,7 == 10 nybbles


void payload(void)
{
  
// Serial.print("millis ");  
// Serial.println(millis());

 long vcc=readVcc()/100;
// Serial.print("Vcc "); 
// Serial.println(vcc,DEC);

 writeBCD(OregonMessageBuffer,6,2,vcc);
 
// Serial.print("Core temp ");
// Serial.println(readTemp(),DEC);
 
 read_sensor();

 read_temp();

 transmitter_send();

 if(!mode_slow)
 {
  pinMode(pin_led, OUTPUT);
  digitalWrite(pin_led, LOW);
  delay(100);
  digitalWrite(pin_led, HIGH);
  delay(100); 
  pinMode(pin_led, INPUT);  
  Serial.flush();
 }
}  



void power_down(void)
{
  

  
}

void power_up(void)
{
 powerup_temp();
 powerup_sensor();

 transmitter_setup();

 delay(1000); // temp sensor needs a second before it can be read
}


void setup() 
{
 Serial.begin(9600);        // connect to the serial port
// Serial.println("Start");
// Serial.flush();
}



void loop() 
{
 int i; 

 power_down();

 for(i=0;i<mode_speed;i++)
 {
  LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);
 }

 power_up();

 payload();

 if(mode_count++==mode_switch)
 {
  mode_speed=mode_slowspeed;
  mode_slow=1;
 }
}


